#!/usr/bin/env python3
from setuptools import find_packages, setup

INSTALL_REQUIRES = []
TESTS_REQUIRE = []

setup(
    name='franklab_mountainsort',
    version='0.1.0.dev0',
    license='',
    description=(''),
    author='',
    author_email='',
    url='',
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
)
